from ._inject import add_value, add_values, set_values, set_values_from_config
from ._inject import assemble, component, factory, value

name = "enterprython"

__author__ = "Tobias Hermann"
__copyright__ = "Copyright 2018, Tobias Hermann"
__email__ = "editgym@gmail.com"
__license__ = "MIT"
